#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <cuda_runtime.h>
#include <cublas_v2.h>

#define	ITER	6

#if TYPE_DOUBLE
	#define	TYPE	double
#else
	#define	TYPE	float
#endif

#include "test-common.h"
#include "cuda.h"

// macro for the legacy cublas
#define	cublasAlloc(a,b,c)	cudaMalloc( (c),(a)*(b) )
#define	cublasFree(a)		cudaFree( a )

int
main(int argc, char *argv[])
{
	TYPE	alpha = (TYPE)3.0;
	TYPE	beta  = (TYPE)0.0;//4.0;
	TYPE	*A, *B, *C, *D;
	TYPE	*dA, *dB, *dC, *dD;
	int	N, M, K;
	int	LDA, LDB, LDC;
	int	i;
	char	mode;
	int	blk;
	FILE	*fp;
	int	id = get_gpu_id();


#if __isDOUBLE__
	enum D_FUNCS	func = GEMV_N;
#else
	enum S_FUNCS	func = GEMV_N;
#endif

	char	*func_name = "";

	cudaSetDevice( id );

	cublasStatus_t stat;
	cublasHandle_t handle;
	cublasAtomicsMode_t atomic_mode = CUBLAS_ATOMICS_ALLOWED;
	stat = cublasCreate( &handle );
	cublasSetAtomicsMode( handle, atomic_mode );

{
	char *argv0 = argv[0];
	if ( strrchr(argv0, '/') != argv0 ) {
		argv0 = strrchr(argv0,'/')+1;
	}
	if ( 0 == strcmp(argv0,TEST_GEMV_N) ) {
		func      = GEMV_N;
		func_name = NAME_GEMV_N;
	}
	if ( 0 == strcmp(argv0,TEST_GEMV_T) ) {
		func      = GEMV_T;
		func_name = NAME_GEMV_T;
	}
	if ( 0 == strcmp(argv0,TEST_SYMV_U) ) {
		func      = SYMV_U;
		func_name = NAME_SYMV_U;
	}
}

	if ( argc > 1 ) {
		fp=fopen(argv[1], "r");
	} else {
		fp=fopen("IN", "r");
	}
	if ( fp == NULL ) {
		fprintf(stderr,"Cannot open parameter file.\n");
		exit(1);
	}

	fscanf(fp,"%d", &N);
	if ( N == 0 ) {
		mode = 'U';
	} else {
		mode= 'U';
	}

	if ( argc > 2 ) {
		sscanf(argv[2], "%d", &blk);
		printf("% BLK specified=%d\n", blk);
	} else {
		blk = -1; //64;
	}


	print_head( func_name );



	int N_fault=0;

while (1) {
	fscanf(fp,"%d", &N);

	if ( N <= 1 ) break;

        int     M = N;
#if 1
if ( argc <= 3 )
//if ( 0 )
{
if ( sizeof(TYPE) == 8 ) {
	if ( M < 9000 ) M = 9000;
} else {
	if ( M < 12600 ) M = 12600;
}
}
#endif
        int     LDM = ((M-1)/(256/sizeof(TYPE))+1);
        LDM *= (256/sizeof(TYPE));

	LDA = ((N-1)/(256/sizeof(TYPE))+1);
	LDA *= (256/sizeof(TYPE));

//	cudaMallocHost((void **)&A, sizeof(TYPE) *(N*LDA));
	cudaMallocHost((void **)&A, sizeof(TYPE) *(M*LDM));
	cudaMallocHost((void **)&B, sizeof(TYPE) *(N*1));
	cudaMallocHost((void **)&C, sizeof(TYPE) *(N*1));
	cudaMallocHost((void **)&D, sizeof(TYPE) *(N*1));

//	cublasAlloc(N*LDA, sizeof(TYPE), (void**)&dA);
	cublasAlloc(M*LDM, sizeof(TYPE), (void**)&dA);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dB);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dC);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dD);

	if(dA==NULL||dB==NULL||dC==NULL||dD==NULL) {
		fprintf(stderr, "# Fail to allocate memory %x %x %x %x\n",
			dA, dB, dC, dD);
		if(dA!=NULL) cublasFree(dA);
		if(dB!=NULL) cublasFree(dB);
		if(dC!=NULL) cublasFree(dC);
		if(dD!=NULL) cublasFree(dD);
		N_fault++;
		goto FFF;
	} else {
		N_fault=0;
	}

	printf("N= %d ", N);
	fflush(stdout);


	for(int i=0;i<N;i++) {
		for(int j=0;j<N;j++) {
			A[j+i*LDA]=(TYPE)(0);
		}
	}
	for(i=0;i<N*1;i++) B[i]=(TYPE)1;
	for(i=0;i<N*1;i++) C[i]=(TYPE)0;
	for(i=0;i<N*1;i++) D[i]=(TYPE)0;

#define	BB		(7)

	if (func==GEMV_N) {
		for(int kk=1;kk<=BB;kk++) {
			for(int k=0;k<N;k++) {
				int j=rand()%N; int i=k;
				A[i+j*LDA] += (TYPE)kk*kk*N;
			}
		}
	}
	if (func==GEMV_T) {
		for(int kk=1;kk<=BB;kk++) {
			for(int k=0;k<N;k++) {
				int i=rand()%N; int j=k;
				A[i+j*LDA] += (TYPE)kk*kk*N;
			}
		}
	}
	if (func==SYMV_U) {
		int	*col = (int *)malloc(sizeof(int)*N);
		for(int kk=1;kk<=BB;kk++) {
			for(int k=0;k<N;k++) {
				col[k] = k;
			}
			for(int k=0;k<N;k++) {
				int i1=rand()%N;
				int i2=rand()%N;
				int j=col[i1]; col[i1]=col[i2]; col[i2]=j;
			}
			for(int k=0;k<N;k++) {
				int i=k, j=col[i];
				A[i+j*LDA] += (TYPE)kk*kk*N;
				A[j+i*LDA] += (TYPE)kk*kk*N;
			}
		}
		free(col);
		for(int i=0;i<N;i++) {
			for(int j=i+1;j<N;j++) {
				A[j+i*LDA] = (TYPE)0;
			}
		}
	}

	for(int k=0;k<N/BB;k++) {
		while ( 1 ) {
		int j=rand()%N;
			if ( B[j] == (TYPE)1 ) {
				B[j] = (TYPE)0;
			break;
			}
		}
	}
	for(int k=0;k<N/BB;k++) {
		while ( 1 ) {
			int j=rand()%N;
			if ( B[j] == (TYPE)1 ) {
				B[j] = (TYPE)(-1);
				break;
			}
		}
	}

//	cublasSetMatrix(N, N, sizeof(TYPE), A, LDA, dA, LDA);
	cublasSetMatrix(LDM, M, sizeof(TYPE), A, LDM, dA, LDM);
	cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
	cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
	cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);

	cudaThreadSynchronize();

double	tt[1000];
double	t1 = get_wtime();

	for(i=0;i<ITER;i++){

	cudaThreadSynchronize();
	double	tt1 = get_wtime();

	int	incb = 1;
	int	incc = 1;
	char	*opt = "U";

	switch (func) {
	case GEMV_N:
                cublasgemv
                        ( handle, CUBLAS_OP_N, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
		break;
	case GEMV_T:
                cublasgemv
                        ( handle, CUBLAS_OP_T, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
		break;
	case SYMV_U:
                cublassymv
                        ( handle, CUBLAS_FILL_MODE_UPPER, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
		break;
	}

	cudaThreadSynchronize();
	double	tt2 = get_wtime();
	tt[i]=tt2-tt1;

	}
EEE:;

double	t2 = get_wtime();
	cudaThreadSynchronize();
double	t3 = get_wtime();

	{
	for(i=0;i<ITER;i++) 
	for(int j=i+1;j<ITER;j++)
	if(tt[i]>tt[j]){ double s=tt[i]; tt[i]=tt[j]; tt[j]=s; }
	}

	double	tmax=0.0;
//	for(i=0;i<ITER;i++) tmax=fmax(tt[i],tmax);
	tmax=tt[ITER-1];

	double	ttotal=0.0;
	for(i=0;i<ITER;i++) ttotal+=tt[i];
	ttotal-=tmax;
//double	t = (ttotal)/(ITER-1);
double	t = tt[(ITER-1)/2];

	double	tvar=0.0;
	for(i=0;i<ITER;i++) tvar+=(tt[i]-t)*(tt[i]-t);
	tvar-=(tmax-t)*(tmax-t);
	tvar=sqrt(tvar)/(ITER-1);

	fflush(stdout);
	printf("%g [s] ", t);
	printf("%g GFLOPS ", ((2.0*N)*N/t*1e-9));
	printf("%g ", tvar);
	if ( tvar/t > 2E-2 ) {
		printf("**%g**", tvar/t );
	}
	printf("%d ", LDA);
	fflush(stdout);
GGG:
	fflush(stdout);

	cublasFree(dA); //printf(":dA");
	cublasFree(dB); //printf(":dB");
	cublasFree(dC); //printf(":dC");
	cublasFree(dD); //printf(":dC");

FFF:
	cudaFreeHost(A); //printf(":A");
	cudaFreeHost(B); //printf(":B");
	cudaFreeHost(C); //printf(":C");
	cudaFreeHost(D); //printf(":D");

	printf("\n");
	fflush(stdout);

	if ( N_fault > 4 ) break;
	if ( feof(fp) ) break;
}
	cublasDestroy( handle );

	fclose(fp);
}

