#include "double.h"
#include "aspen_devel.h"

#include "../tuning/CURRENT_GPU"


extern "C" void
API(DSYMVu_LARGE) ( const int blk,
         const int n, const scalar_t * a, const int lda,
         const scalar_t * x, scalar_t * y,
         const scalar_t alpha, const scalar_t beta )
{

// y = alpha * A * x + beta * y
// y(1:n1)
// x(1:n2)

   if ( n <= 0 ) return;
   if ( alpha == CONST(0) && beta == CONST(1) ) return;

   int Ver = API(get_device_CC) ( );

   switch ( Ver ) {
#if CURRENT_GPU==Tesla
   case Tesla:
      API(DSYMVu_LARGE_TESLA)
         ( blk, n, a, lda, x, y, alpha, beta );
      break;
#endif
#if CURRENT_GPU==Fermi
   case Fermi:
      API(DSYMVu_LARGE_FERMI)
         ( blk, n, a, lda, x, y, alpha, beta );
      break;
#endif
#if CURRENT_GPU==Kepler
   case Kepler:
      API(DSYMVu_LARGE_KEPLER)
         ( blk, n, a, lda, x, y, alpha, beta );
      break;
#endif
#if CURRENT_GPU==Kepler2
   case Kepler2:
      API(DSYMVu_LARGE_KEPLER2)
         ( blk, n, a, lda, x, y, alpha, beta );
      break;
#endif
#if CURRENT_GPU==Maxwell
   case Maxwell:
      API(DSYMVu_LARGE_MAXWELL)
         ( blk, n, a, lda, x, y, alpha, beta );
      break;
#endif
   default:
	break;
   }

}

