#!/bin/bash

prolog()
{
	if [ x${CUDA_PATH} = x ]; then
		echo "Please set up the CUDA_PATH variable."
		echo "(1) export CUDA_PATH=/usr/local/cuda"
		echo "(2) setenv CUDA_PATH /usr/local/cuda"
		exit 1
	fi
	AT_WORK_DIR=logs
	AT_LOG_FILENAME=sgemvn-at-log
	AT_OPT_FILENAME=sgemvn-at-opt.c
	AT_OPT2_FILENAME=sgemvn-at-opt2.h
	AT_OPT_EXE=sgemvn-at-pred
	AT_GP_PLOT=aa.plot
	BENCHCODE=../bench/test2-sgemv-n
	IN_DATA=IN-a

	if [ x${HOST} = x ]; then
		HOST=`/bin/hostname`
	fi

	/bin/sh ./get_dev_info.sh > DEV_INFO

	DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
	MP=`awk '/MP=/{ print $2 }' DEV_INFO`
	CG=`awk '/CG=/{ print $2 }' DEV_INFO`
	MAXDIM=`awk '/MAXDIM2/{ print $2 }' DEV_INFO`

	\rm DEV_INFO

	if [ -e CURRENT_GPU ]; then
		\rm CURRENT_GPU
	fi
	echo '#define CURRENT_GPU '$CG > CURRENT_GPU

	if [ $CG -eq 130 ]; then
		MAX_REGS=32768
		MAX_BLK=8
		SHMEM_UNIT_SIZE=512
	fi
	if [ $CG -eq 200 ]; then
		MAX_REGS=32768
		MAX_BLK=8
		SHMEM_UNIT_SIZE=128
	fi
	if [ $CG -eq 300 ]; then
		MAX_REGS=65536
		MAX_BLK=16
		SHMEM_UNIT_SIZE=256
	fi
	if [ $CG -eq 350 ]; then
		MAX_REGS=65536
		MAX_BLK=16
		SHMEM_UNIT_SIZE=256
	fi
	if [ $CG -eq 500 ]; then
		MAX_REGS=65536
		MAX_BLK=32
		SHMEM_UNIT_SIZE=256
	fi

	CORES=`cat /proc/cpuinfo | awk 'BEGIN{c=1}/cores/{c=$4}END{print c}'`
	if [ $CORES -gt 1 ]; then
		EXEC='' #'/usr/bin/numactl --physcpubind=0 '
	else
		EXEC=''
	fi

	FIT_INIT='o0=10; o1=1; c0=1; c1=1; c2=1;' 

	ALL_BLOCK_X_LIST=" \
		100006 100007 100008 100009 100010 \
		100011 100012 100013 100014 100015 \
		100016 \
		100106 100107 100108 100109 \
 		1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 \
 		101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 \
 		201 202 203 204 205 206 207 208 209 210 211 212 \
		1001 1002 1003 \
		1101 1102 1103 \
		2001 2002 2003 2004 2005 2006 2007 2008 \
		2101 2102 2103 2104 2105 2106 \
		3001 \
		3101 \
		4001 4002 4003 4004 4005 \
		5001 \
		6001 6002 6003 6004 \
		7001 \
		8001 8002 8003 \
		9001 \
		10001 10002 \
		11001 \
		12001 12002 \
		13001 \
		14001 14002 \
		16001 \
		18001 \
		20001 \
		22001"

	BLOCK_X_LIST=$ALL_BLOCK_X_LIST

	if [ x${DEVICE} = xunknown ]; then
		echo '**** CAUTION !!! ****'
		echo 'CUDA environment does not match to your GPU'
		echo 'Check the installed NVIDIA driver and the version of CUDA!'
		exit 1
	fi

	/bin/sh ./touch_header.sh

	if [ -e $AT_WORK_DIR ]; then
		\rm -rf $AT_WORK_DIR
	fi
}

echo_Message()
{
	echo '<--/****************************************'
	echo 'Automatic Performance Tuning for SGEMV -N'
	date
	echo -n 'Host on '; echo `hostname`
	echo 'Device is '$DEVICE
	if [ x${1} = xbegin ]; then
		echo 'Start.'
	fi
	if [ x${1} = xend ]; then
		echo 'Successfully completed.'
	fi
	if [ x${1} = xterminate ]; then
		echo 'Terminated unfortunately.'
	fi
	echo '****************************************/-->'
}

update_ASPEN()
{
	cd ../src
	\rm sgemv_n.cu_o
	\rm sgemv_n-*.cu_o
	make -f $1
	cd ../bench
	\rm test-f.o test2-f.o
	make test-s
	cd ../tuning
}

init_IN_file()
{
	echo 1 > $IN_DATA
	for N in \
		50 100 150 200 250 300 350 400 500 600 700 \
		800 900 1000 1200 1400 1600 1900 2300 2800
	do
		echo  $N >> $IN_DATA
	done
	N=3500
	NN=$MAXDIM
#	NS=`expr 56 "*" $MP`
	NS=`expr 83 "*" $MP`
	while [ $N -lt $NN ]
	do
		echo $N >> $IN_DATA
		N=`expr $N "+" $NS`
	done
		echo $MAXDIM >> $IN_DATA	
	echo 0 >> $IN_DATA
}

init_AT_API()
{
	for BLOCK_X in $BLOCK_X_LIST
	do
		echo '#define KERNEL_'$BLOCK_X' 1'
	done
}

do_AT_sampler1()
{
	echo 1 > IN-1
	echo 6000 >> IN-1
	echo -1 >> IN-1

	echo '<--/****************************************'
	echo 'Check the number of registers and shared memory'
	echo 'and Screen Kernels by occupancy.'
	echo '****************************************/-->'

	\touch sgemvn_opt4.h
	\rm    sgemvn_opt4.h
	\touch sgemvn_opt5.h
	\rm    sgemvn_opt5.h

	echo '#if 0' > sgemvn_opt4.h
	echo_Message sampler1 >> sgemvn_opt4.h
	echo '#endif' >> sgemvn_opt4.h
	echo '#undef AT_VERSION' >> sgemvn_opt4.h


        /bin/sh ./get_reguse.sh $CG \
                ../src/sgemv_n-X_template \
                ../src/sgemv_n-X_template \
                ../src/sgemv_n-X_template \
                ../src/sgemv_n-X_template \
                ../src/sgemv_n-X_template \
                small_kernel > log-reg-all


	for BLOCK_X in $BLOCK_X_LIST
	do

# ============================================================

#		BASE=`expr $BLOCK_X "/" 2000`
#		BASE=`expr $BASE "+" 1`
#		BASE=`expr $BASE "*" 32`
		if [ $BLOCK_X -lt 100000 ]; then
		BASE=`expr $BLOCK_X "/" 1000`
		BASE=`expr $BASE "+" 2`
		BASE=`expr $BASE "*" 16`
		else
		BASE=16
		fi
		BLOCK_Y=`expr $BLOCK_X "%" 100`

		cat log-reg-all | awk '{if(($2=='$BASE')&&($3=='$BLOCK_Y'))print $1}' > CX
		CX=`cat CX`
		if [ x${CX} = "x" ]; then
			CX=1
		fi

# ============================================================

		M1=`expr $BLOCK_Y "*" $BASE`

		M=`expr $BLOCK_Y "/" 2`
		M=`expr $M "*" 2`
		M=`expr $M "+" 1`
		M=`expr $M "*" $BASE`
		M2=`expr $M "*" 4`

		M=`expr $BLOCK_X "/" 100`
		M3=`expr $M "%" 10`

		MM_PAT=`/bin/sh ./get_mult.sh $M1 $CX $M2 $CG`

		M=`expr $M3 "+" 1`
		\touch log_MM
		\rm log_MM
		for MM in $MM_PAT
		do
			echo $MM >> log_MM
		done
		MM=`cat log_MM | \
			tail -$M | head -1 `
		MULTIPLICITY=`echo $MM | \
			awk '{ gsub(/\(.*\)/,""); print; }'`
		OO=`echo $MM | \
			awk '{ gsub(/[\(\)]/," "); if($2==100)$2=99; print $2; }'`
		\rm log_MM

		M=`expr $MP "*" $BASE`
		CYCLE=`expr $M "*" $MULTIPLICITY`

		echo '' >> sgemvn_opt4.h
		echo '// ID['$BLOCK_X'] ('$BASE','$BLOCK_Y') * '$MULTIPLICITY >> sgemvn_opt4.h
		echo '#define _BLOCKSIZE_'$BLOCK_X' '$BASE >> sgemvn_opt4.h
		echo '#define _UX_'$BLOCK_X' '$BLOCK_Y >> sgemvn_opt4.h
		echo '#define _MX_'$BLOCK_X' '$M2 >> sgemvn_opt4.h
		echo '#define _CX_'$BLOCK_X' '$CX >> sgemvn_opt4.h
		echo '#define _MULTI_'$BLOCK_X' '$MULTIPLICITY >> sgemvn_opt4.h
		echo '#define _OCCUP_'$BLOCK_X' '$OO >> sgemvn_opt4.h
		echo '#define _CYCLE_'$BLOCK_X' '$CYCLE >> sgemvn_opt4.h

# ============================================================

		echo -e 'Kernel ID:\t'$BLOCK_X'\tRegs:\t'$CX'\tMULTI:\t'$MULTIPLICITY'\tCYCLE:\t'$CYCLE'\t/ '$OO | tee -a sgemvn_opt5.h

	done

	echo '' >> sgemvn_opt4.h
	echo '#if defined(__CY__)' >> sgemvn_opt4.h
	echo '#undef __CY__' >> sgemvn_opt4.h
	echo '#endif' >> sgemvn_opt4.h
	echo '#define __CY__' >> sgemvn_opt4.h

	if [ $CG -le 200 ]; then
		O_LIMIT=15
	else
		O_LIMIT=20
	fi

	cat sgemvn_opt5.h | \
		sort -r -k 11 | sort -r -k 9 | \
		awk 'BEGIN{ \
			; \
		} \
		{ \
			multi = $7; \
			cycle = $9; \
			occup = $11; \
			if ( cycle != CYC ) { \
				CYC = cycle; \
				OO = occup; \
				NN = 0; \
			} \
			if ( occup != OO ) { \
				if ( NN == 0 ) { \
					NN = occup; \
				} \
			} \
			if ( multi < 2 || occup < '$O_LIMIT' || \
				( occup != OO && occup != NN && \
					OO*.75 > occup ) ) { \
				print $3; \
			} \
		} \
		END{ \
			;
		}' | \
		tee skip_list | \
		awk '{ print "Sampling for Kernel "$1" will be skipped" }'

	cat skip_list | \
		awk 'BEGIN{ \
			print ""; \
		} \
		{ \
			print "#undef _OCCUP_"$1; \
			print "#define _OCCUP_"$1" 0"; \
		} \
		END{ \
			print ""; \
		}' >> sgemvn_opt4.h

	echo '*/' >> sgemvn_opt5.h
	cat sgemvn_opt4.h >> sgemvn_opt5.h

	echo '/*' > sgemvn_opt4.h
	cat sgemvn_opt5.h >> sgemvn_opt4.h

	\rm sgemvn_opt5.h

	\rm IN-1

}

do_AT_sampler()
{
	if [ ! -d $AT_WORK_DIR ]; then
		mkdir $AT_WORK_DIR
	fi

	echo '<--/****************************************'
	echo 'Enter the Sampling Phase, it may take a while.'
	echo '****************************************/-->'

	for BLOCK_X in $BLOCK_X_LIST
	do

#		BASE=`expr $BLOCK_X "/" 2000`
#		BASE=`expr $BASE "+" 1`
#		BASE=`expr $BASE "*" 32`
		if [ $BLOCK_X -lt 100000 ]; then
		BASE=`expr $BLOCK_X "/" 1000`
		BASE=`expr $BASE "+" 2`
		BASE=`expr $BASE "*" 16`
		else
		BASE=16
		fi
		BLOCK_Y=`expr $BLOCK_X "%" 100`

		cat sgemvn_opt4.h | \
		awk '/define[\t ]+_CYCLE_'$BLOCK_X'[\t ]+/{ \
			gsub(/\(/,""); gsub(/\)/,""); \
			if($3*1>=0){ print $3; exit } \
		}' > CY
		CY=`cat CY`

		cat sgemvn_opt4.h | \
		awk '/define[\t ]+_OCCUP_'$BLOCK_X'[\t ]+/{ \
			gsub(/\(/,""); gsub(/\)/,""); \
			if($3*1>=0){ print $3 } \
		}' | tail -1 > OO
		OO=`cat OO`


		echo 'Sampling for Kernel '$BLOCK_X
		echo 'CYCLE= '$CY

		touch $IN_DATA--a
		\rm $IN_DATA--a

		cat $IN_DATA | \
			awk '/^[0-9]/{ \
				if ( $1 < 1 ) { \
					N=CY='$CY'; \
					MAXDIM='$MAXDIM'; \
					count=0; \
					for(N=CY;N-32<'$MAXDIM';N+=CY) { \
					count++; \
			if ( count < 8 ) { \
				if(N-32<=MAXDIM)print N-32; \
			} \
			if ( count < 12 ) { \
				if(N-16<=MAXDIM)print N-16; \
			} \
			if ( count < 16 ) { \
				if(N-04<=MAXDIM)print N-04; \
			} \
			if ( count < 20 ) { \
				if(N-01<=MAXDIM)print N-01; \
				if(N   <=MAXDIM)print N   ; \
			} \
			if ( count < 16 ) { \
				if(N+04<=MAXDIM)print N+04; \
			} \
			if ( count < 12 ) { \
				if(N+16<=MAXDIM)print N+16; \
			} \
			if ( count < 8 ) { \
				if(N+32<=MAXDIM)print N+32; \
			} \
					} \
					exit;
				} \
			} \
			{ print; }' > $IN_DATA--a

		cat $IN_DATA--a | \
			awk '{ printf("%06d\n", $0); }' | \
			sort | \
			uniq | \
			awk '{ print $0*1; } END{ print 0; }' > $IN_DATA--b
			

		if [ $CG -le 200 ]; then
			O_LIMIT=15
		else
			O_LIMIT=20
		fi

		if [ $OO -lt $O_LIMIT ]; then

		echo 'Kernel ID: '$BLOCK_X' is skipped due to the Low Occupancy'
		cat $IN_DATA--b | \
			awk 'BEGIN{ getline } \
				{ if ( $1 < 1 ) exit; \
					print "N= "$1" 0 [s] 0 GFLOPS 0 0" } \
			' > $AT_WORK_DIR/log-$BLOCK_X

		else

		$EXEC $BENCHCODE $IN_DATA--b $BLOCK_X | \
			awk '/^N=/{print; fflush(); }' | \
			tee $AT_WORK_DIR/log-$BLOCK_X

		fi

		cp $IN_DATA--b $AT_WORK_DIR/$IN_DATA--$BLOCK_X

		\rm $IN_DATA--a $IN_DATA--b


		echo 'CYCLE= '$CY | tee -a logs/log-$BLOCK_X

	done
}

epilog()
{
	if [ -d $AT_WORK_DIR ]; then
		LOG=log-sgemvn-AT1X.`date +%Y.%m.%d.%T`
		mv $AT_WORK_DIR $LOG
		cp ../src/sgemv_n-X_template.cu $LOG
		cp ../template//gemv_n-small_template.incl $LOG
		mv $AT_LOG_FILENAME $LOG
	fi
	if [ -f $IN_DATA ]; then
		\rm $IN_DATA
	fi
	if [ -f $AT_OPT_EXE ]; then
        	\rm $AT_OPT_EXE
	fi
	if [ -f fit.log ]; then
		\rm fit.log
	fi
	if [ -f SdkMasterLog.csv ]; then
		\rm SdkMasterLog.csv
	fi
	if [ -f deviceQuery.txt ]; then
		\rm deviceQuery.txt
	fi
}



prolog

echo_Message begin

if [ x${CUDA_CHECK} = xFAILED ]; then
	echo '*** WARNING ***'
	echo 'CUDA_SDK_PATH does not match the one specified on sgemvn.sh'
	echo 'PLEASE correct CUDA_SDK_PATH.'
	exit 1
fi

if [ $CG -lt 200 ]; then

	touch $AT_LOG_FILENAME

	generate_AT_API > $AT_OPT_FILENAME

	update_ASPEN Makefile.at

	echo '*** CAUTION ***'
	echo 'Because GT200 core uses a propriately tuned routine for SGEMV-N,'
	echo 'NO tuning process is required at the moment.'

	epilog

	echo_Message end

	exit 0;
fi

init_IN_file

init_AT_API >$AT_OPT2_FILENAME

touch sgemvn_opt4.h
\rm sgemvn_opt4.h
touch sgemvn_opt4.h

update_ASPEN Makefile

do_AT_sampler1

update_ASPEN Makefile

do_AT_sampler

cd logs/
touch sgemvn-auto2.h
\rm  sgemvn-auto2.h

/bin/sh ../bmvns.sh
/bin/sh ../cmvns.sh
cat sgemvn-auto.h
/bin/sh ../dmvns.sh

cat sgemvn-auto2.h
cp  sgemvn-auto2.h ../
cd ../

update_ASPEN Makefile.at

epilog

echo_Message end

touch .done-sgemvn

exit 0


